/*  Copyright (C) MOXA Inc. All rights reserved.

    This software is distributed under the terms of the
    MOXA License.  See the file COPYING-MOXA for details.
*/
/*
    driver_lx.c

    This defines driver entry functions (open, read, write, close) for Linux.

    2009-10-15	CF Lin
		new release
*/
#include "driver.h"
#include <linux/cdev.h>


/*	the ISR, it is a functin, not a thread 
	Inputs:
		<irq> irq #, useless in this case
		<dev_id> harware port
	Returns:
		IRQ_HANDLED
*/
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,18) 
static irqreturn_t driver_interrupt( int irq, void *dev_id)
#else
static irqreturn_t driver_interrupt( int irq, void *dev_id, struct pt_regs *notused)
#endif
{
	HWPORT *port = (HWPORT*) dev_id;

	/* 
	 * Note: fix the Problem IER be enabled but port is not used 
	 * 1. in port initial, IER should = 0
	 * 2. when port open (port->num_uses > 0), IER enable
	 * 3. so we only check open port for interrupt
	 */
	 
	if (port->num_users) {
		if (driver_port_interrupt(port)) /* return the type of interrupt, otherwise return 0 */
		{
			#ifndef X86
			if ( port->port_idx == 1 )
			{
				/* prevent port 2 from poping up interrupts continuously */
				__raw_writel(1, EP93XX_GPIO_B_INT_ACK);
			}
			#endif 
			return IRQ_HANDLED;
		}
		else {
			return IRQ_NONE; /* for share irq, it means this is not a interrupt i cause */
		}
	}
	return IRQ_NONE;
}





/*	close the access to a port for a user 
	Inputs:
		<inode> representing the index of the hardware port
		<file>	representing the user
	Returns:
		on success, return 0.
*/
int 
xxx_close ( struct inode *inode, struct file *file )
{
	HWUSER *user = (HWUSER*) file->private_data;

	(void) inode;

	dbg_function_in(); /* debug */
	/* remove a user from the pool */
	if (driver_port_remove_user(user)) 
		return -EINVAL;
	dbg_function_out(); /* debug */
    return 0;
}

/*	open an access to a port for a user 
	Inputs:
		<inode> representing the index of the hardware port
		<file>	representing the user
	Returns:
		on success, this function returns 0, 
		otherwise it returns a negative value.
*/
int 
xxx_open( struct inode *inode, struct file *file ) 
{
	HWPORT *port;
	HWUSER *user;
	int indx = MINOR(inode->i_rdev);

	dbg_function_in(); /* debug */
	port = driver_port_lookup(indx);
	if (port == NULL) 
		return -EIO;
	/* allocate memeory for a port user */
	user = driver_port_create_user(port);
	if (user==NULL)
		return -ENOMEM;
	file->private_data = (void *) user;
	dbg_function_out(); /* debug */
    return 0;
}

/*	read data into a user buffer from a port for a user 
	Inputs:
		<file>	representing the user
		<buf> point to the data buffer
		<max> the capacity of the buffer
		<loff> offset, useless
	Returns:
		on success, this function returns 0, 
		otherwise it returns a negative value.
*/
ssize_t 
xxx_read( struct file *file, char *buf, size_t max, loff_t *loff )
{
	HWUSER *user;
	HWPORT *port;
	unsigned int len, blocking;
	size_t bytes;
	(void) loff;

	dbg_function_in(); /* debug */
	
	/* 
	 * XXX: read size should be the 20*n, 
	 * because this driver only can write 20*n 
	 * see hwport_write for detail
	 */
	max = (max/20)*20;

	user = (HWUSER*) file->private_data;
	if (!user || (port = user->hwport)==NULL)
	{
	   return -EINVAL;
	}
	if( !access_ok( VERIFY_WRITE, buf, max )) 
	{
	   return -EINVAL;
	}
	blocking = (file->f_flags & O_NONBLOCK)? 0:1;
	bytes = 0;
	while(user->running && max)
	{
		/* read data from rx_buffer/hardware */
		len = driver_port_copy_to_user(port, buf, max);
		if (len)
		{
			/* move pointers */
            max -= len;
            buf += len;
			bytes += len;
		}
		else if (bytes)
		{
			break;
		}
		else if (blocking)/* in blocking mode */
		{
			if (wait_event_interruptible(port->wait_rx_event, ISBUFDATATHERE(port->rx_buf)))
				return -ERESTARTSYS; /* let uplayer FS to handle it */
			/* else, data is there */
		}
		else
		{
			break; /* NON_BLOCK mode */
		}
	}
DEBUGMSG("port[%d]: %d bytes read of %d max\r\n", port->port_idx, bytes, max);
	dbg_function_out(); /* debug */
    return bytes;
}

/*	write buffed data onto a port for a user 
	Inputs:
		<file> representing the user
		<buf>  point to the data buffer
		<max>  the length of data
		<loff> offset, useless
	Returns:
		on success, this function returns 0, 
		otherwise it returns a negative value.
*/
ssize_t
xxx_write( struct file *file, const char *buf, size_t max, loff_t *loff )
{
	HWUSER *user;
	HWPORT *port;
	unsigned int len, blocking;
	size_t bytes;
	char *ptr = (char*) buf;

	/* 
	 * XXX: write size should be the 20*n, 
	 */
	max = (max/20)*20;

	dbg_function_in(); /* debug */
	user = (HWUSER*) file->private_data;
	if (!user || (port = user->hwport)==NULL)
	{
	   return -EINVAL;
	}
	if(!access_ok( VERIFY_READ, buf, max )) 
	{
	   return -EINVAL;
	}
	blocking = (file->f_flags & O_NONBLOCK)? 0:1;
	bytes = 0;
	while(user->running && max)
	{
		/* write data to the TX buffer */
		len = driver_port_copy_from_user(port, ptr, max);
		if (len)
		{
			/* move pointers */
            max -= len;
            ptr += len;
			bytes += len;
		}
		else if (bytes)
		{
			break;
		}
		else if (blocking)/* in blocking mode */
		{
			if (wait_event_interruptible(port->wait_tx_event, ISBUFAVAILABLE(port->tx_buf)))
				return -ERESTARTSYS; /* let uplayer FS to handle it */
		}
		else
		{
			break; /* NON_BLOCK mode */
		}
	}
	dbg_function_out(); /* debug */

    return bytes;
}

int 
xxx_ioctl( struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg )
{
	HWUSER *user;
	HWPORT *port;
	int ret = 0;
	unsigned long flags;

	(void) inode;

	dbg_function_in(); /* debug */
	user = (HWUSER*) file->private_data;
	if (!user || (port = user->hwport)==NULL)
	{
		return -EINVAL;
	}
	local_irq_save(flags);
DEBUGMSG("xxx_ioctl : %d\n", cmd);
	/* handle common commands */
	ret = driver_port_common_ioctl(user, cmd, (void*) arg, 0, (void*) arg, 0, NULL);
	if (ret == -99)
	{
		ret = hwport_ioctl(port, cmd, (void*) arg, 0, (void*) arg, 0, NULL);
		if (ret) 
			ret = -EIO;
	}
	local_irq_restore(flags);
	dbg_function_out(); /* debug */
    return ret;
}

int 
xxx_fasync( int fd, struct file *file, int count ) /* inode, file, count */
{
	int retval = 0;
    return retval;
}

unsigned int 
xxx_select( struct file *file, struct poll_table_struct *wait )
{
	HWUSER *user = (HWUSER*) file->private_data;
	HWPORT *port;
	unsigned int n, mask = 0;

	dbg_function_in(); /* debug */
	port = user->hwport;
    /* every event queue that could wake up the process
     * and change the status of the poll operation
     * can be added to the poll_table structure by
     * calling the function poll_wait:  
     */
    /*     _select para, wait queue, _select para */
    poll_wait(file, &port->wait_rx_event, wait);
    poll_wait(file, &port->wait_tx_event, wait);

	driver_port_get_rx_bytes(port, n); /* macro */
    if( n >= MSG_UNIT_SIZE ) 
	{
		/*	fifo has some telegrams */
		/*	Return a bit mask describing operations that could be 
			immediately performed without blocking.
		*/
		/*	POLLIN This bit must be set if the device can be read 
			without blocking. 
			POLLRDNORM This bit must be set	if "normal'' data is 
			available for reading.
			A readable device returns (POLLIN | POLLRDNORM)
		 */
		mask |= (POLLIN | POLLRDNORM);	/* readable */
    }
	driver_port_get_tx_space(port, n); /* macro */
    if( n >= MSG_UNIT_SIZE ) 
	{
		/*	Return a bit mask describing operations that could be 
			immediately performed without blocking.
		*/
		/*	POLLOUT This bit must be set if the device can be written 
			without blocking. 
			POLLWRNORM This bit must be set
			*/
		mask |= (POLLOUT | POLLWRNORM);	/* writeable */
    }
	dbg_function_out(); /* debug */
    return mask;
}


static struct file_operations xxx_fops = 
{ 
    .owner	=	THIS_MODULE,
    .open	=	xxx_open,
    .release =	xxx_close,
    .read	=	xxx_read,
    .write	=	xxx_write,
    .poll	=	xxx_select,
    .ioctl	=	xxx_ioctl,
    .fasync	=	xxx_fasync,
};


#ifdef PCI_BUS
static int pci_port_counter;
static int has_registered_chrdev;
static void __exit	driver_exit_module(void);
/* 
 * Create table: 
 * vendor, device, subvendor, subdevice, class, class mask, driver_data
 * note: 
w
 *		1. vendor id, device id can be PCI_ANY_ID
 *		2. lspci -vvvn for more detail	
 */
static struct pci_device_id xxx_pci_tbl[]= {
	{0x1393, 0xa000, PCI_ANY_ID, PCI_ANY_ID, 0, 0, 2}, /* 2 is port number */
	{0,}
};
MODULE_DEVICE_TABLE (pci, xxx_pci_tbl);


static void add_cdev_device(int index) {
	int err, devno = MKDEV(DEVICE_DRIVER_MAJOR, index);

	/* create a cdev device */
	struct cdev *c = cdev_alloc();
	c->owner = THIS_MODULE;
	c->ops = &xxx_fops;

	/* add to kernel */
	err = cdev_add(c, devno, 1);
	if (err)
		printk("%s error\n", __FUNCTION__);
}

static struct class *charmodule_class;

static int __devinit xxx_probe(struct pci_dev *pci_dev, const struct pci_device_id *pci_id)
{
	int i;
	int j = 0;
	unsigned long base_addr, flag;
	int ret;
	int num_port;
	dev_t dev;
	IRQMEM irqmem = {};
	HWPORT *port;


	/* initial device */
	if (pci_enable_device(pci_dev))
		return -EIO;
	num_port = pci_id->driver_data;

	/* request a chrdev region */
	if (has_registered_chrdev == 0) {
		dev = MKDEV(DEVICE_DRIVER_MAJOR, 0);
		if ((ret = register_chrdev_region(dev, 4, "can")) < 0) {
			printk("can't alloc a chrdev_region\n");
		}
		has_registered_chrdev = 1;

		/* create udev class */
		charmodule_class = class_create(THIS_MODULE, "EPM3112");
        if(IS_ERR(charmodule_class)) {
            printk("error: failed in creating class.\n");
            ret=-ENOMEM;
			return ret;
        }
	}

	/* read configuration space */
	for (i = 0; i<6; i++) {
		p("bar %d start=%x end=%x ",
				i,
				pci_resource_start(pci_dev, i),
				pci_resource_end(pci_dev,i));

		base_addr = pci_resource_start(pci_dev, i);
		flag = pci_resource_flags(pci_dev,i);

		/* check flag */
		if (flag & IORESOURCE_IO) {
			p("flag=IORESOURCE_IO\n");
		} else if (flag & IORESOURCE_MEM) {
			p("flag=IORESOURCE_MEM\n");
		} else {
			p("flag=IORESOURCE_OTHER leave the probe loop\n");
			break;
		}

		if (j > num_port) {
			printk("%s:  probe pci config space more than num_port=%d\n", __FUNCTION__, num_port);
			return -1;
		}

		/* Initial a port here */
		irqmem.irq_num = pci_dev->irq;
		irqmem.base_addr = base_addr;
		port = driver_port_init(pci_port_counter, 256, irqmem); /* port_idx, mem_size: initialize all informaion for port */
		if (port == NULL)
		{
			ret = -ENOMEM;
			goto err;
		}

		/* Request irq: name should be const char (or should has global scope) */
		sprintf(port->name, "%s%d", DEVICE_DRIVER_NAME, pci_port_counter); /* device node */
		ret = request_irq(port->irq_num, driver_interrupt, IRQF_SHARED, port->name, port);
		if (ret)
		{
			printk("port [%d] : can't request IRQ %d \n", i, port->irq_num);
			goto err;
		}
		p("port [%d]: base_addr (0x%x) data_addr (0x%x)",
		 port->port_idx, (unsigned int) port->base_addr, (unsigned int)port->data_addr);

	
		/* register cdev device */
		add_cdev_device(pci_port_counter);

		/* add device note */
		device_create(charmodule_class, NULL, MKDEV(DEVICE_DRIVER_MAJOR, pci_port_counter), NULL, port->name);

		/* End of initial a port */
		pci_port_counter++;
		j++;

	}
	return 0;

	err:
		driver_exit_module();
		return ret;
}

static void __devexit xxx_remove(struct pci_dev *pci_dev)
{
	int i;
	dev_t dev; 

	/* free irq / region / memory here  */
	driver_port_free_all();

	dev = MKDEV(DEVICE_DRIVER_MAJOR, 0);
	unregister_chrdev_region(dev, 4);

	for (i = 0; i< pci_port_counter; i++)
		device_destroy(charmodule_class, MKDEV(DEVICE_DRIVER_MAJOR, i));

	class_destroy(charmodule_class);
}

//#define PCI_DRIVER_NAME "MOXA_CAN"
static struct pci_driver xxx_pci_driver = {
    .name       = "moxa can",
    .id_table   = xxx_pci_tbl,
    .probe      = xxx_probe,
    .remove     = xxx_remove,
// #ifdef CONFIG_PM
//     .suspend    = xxx_pm_suspend,
//     .resume     = xxx_pm_resume,
// #endif /* CONFIG_PM */
};

#endif 

/* 
 * Note: normally, unregister_chrdev should put in the pci_remove,
 * here is for compatible with local bus driver
 */
static void __exit	
driver_exit_module(void)
{
	dbg_function_in(); /* debug */

#ifdef LOCAL_BUS
	/* free all ports */
	driver_port_free_all();
	/* remove device from the system */
	unregister_chrdev(DEVICE_DRIVER_MAJOR, DEVICE_DRIVER_NAME);
#elif defined PCI_BUS
	/* Note: free all ports in remove() */
	pci_unregister_driver(&xxx_pci_driver);
#endif 

	dbg_function_out(); /* debug */
}



static int __init	
driver_init_module(void)
{
#ifdef LOCAL_BUS
	HWPORT	*port;
	int ret = 0;
	int max_port, i;
	char name[64];

	dbg_function_in(); /* debug */
	/* walk through each port */
	max_port = driver_private_get_maxport();
	for (i=0; i < max_port; i++)
	{
		port = driver_port_init(i, 256);
		if (port == NULL)
		{
			ret = -ENOMEM;
			goto err;
		}
		/*
			int request_irq(unsigned int irq,			// interrupt number  
				  void (*handler)(int, void *, struct pt_regs *),
				  unsigned long irqflags, const char *devname,
				  void *dev_id);

		   dev_id - The device ID of this handler (see below). This parameter is usually 
		   set to NULL, but should be non-null if you wish to do  IRQ  sharing.
		   This doesn't matter when hooking the interrupt, but is required so that, 
		   when free_irq() is called, the correct driver is unhooked. Make sure you
		   pass the same pointer to free_irq().
		*/
		sprintf(name, "%s%d", DEVICE_DRIVER_NAME, i); /* device node */
		ret = request_irq(port->irq_num, driver_interrupt, IRQF_SHARED, name, port);
		if (ret)
		{
			printk("port [%d] : can't request IRQ %d \n", i, port->irq_num);
			goto err;
		}
		else
		{
			/* 
				set_irq_type(xxx_IRQ_PIN, IRQT_LOW); 
			*/
		}
DEBUGMSG("port [%d]: base_addr (0x%x) data_addr (0x%x)\n",
		 port->port_idx, (unsigned int) port->base_addr, (unsigned int)port->data_addr);
	}
	/* register to a character mode device driver on kernel */
	if (register_chrdev(DEVICE_DRIVER_MAJOR, DEVICE_DRIVER_NAME, &xxx_fops))
	{
		ret = -EIO;
		goto err;
	}
	dbg_function_out(); /* debug */
	return 0;
err:
	driver_exit_module();
	return ret;
#elif defined PCI_BUS
	int ret;
	/* register pci driver */
	ret = pci_register_driver(&xxx_pci_driver);
	if (ret) 
		return ret;
	return 0; /* stop here */
#endif
}

module_init(driver_init_module);
module_exit(driver_exit_module);
MODULE_AUTHOR("Victor Yu. <victor.yu@moxa.com>");
MODULE_DESCRIPTION("This is a CAN-Bus Raw driver");
MODULE_LICENSE("GPL");


